#ifndef CONSTS
#define CONSTS

/**************************************************************************** 
 *
 * This header file contains utility constants & macro definitions.
 * 
 ****************************************************************************/
/* Thread scheduling */
#define MAXTHREADS              20
 
/* Messages management */
#define MAXMESSAGES             20
 

/* Hardware & software constants */
#define PAGESIZE		4096	/* page size in bytes */
#define WORDLEN			4		/* word size in bytes */
#define PTEMAGICNO		0x2A


#define ROMPAGESTART	0x20000000	 /* ROM Reserved Page */


/* timer, timescale, TOD-LO and other bus regs */
#define RAMBASEADDR		0x10000000
#define TODLOADDR		0x1000001C
#define INTERVALTMR		0x10000020	
#define TIMESCALEADDR	0x10000024


/* utility constants */
#define	TRUE		1
#define	FALSE		0
#define	ON		1
#define	OFF		0
#define HIDDEN		static
#define EOS			'\0'

#define NULL ((void *)0xFFFFFFFF)


/* vectors number and type */
#define VECTSNUM	4

#define TLBTRAP		0
#define PROGTRAP	1
#define SYSTRAP		2

#define TRAPTYPES	3


/* device interrupts */
#define DISKINT		3
#define TAPEINT 	4
#define NETWINT 	5
#define PRNTINT 	6
#define TERMINT		7

#define DEVREGLEN	4	/* device register field length in bytes & regs per dev */
#define DEVREGSIZE	16 	/* device register size in bytes */

/* device register field number for non-terminal devices */
#define STATUS		0
#define COMMAND		1
#define DATA0		2
#define DATA1		3

/* device register field number for terminal devices */
#define RECVSTATUS  0
#define RECVCOMMAND 1
#define TRANSTATUS  2
#define TRANCOMMAND 3

/* device common STATUS codes */
#define UNINSTALLED	0
#define READY		1
#define BUSY		3


/* device common COMMAND codes */
#define RESET		0
#define ACK			1

#define STATUSMASK             0x000000FF
#define CAUSEMASK              0x000000FF

/* terminal specific address and codes */
#define TERMINAL0ADDR           0x10000250
#define PRINTCHAR               2
#define CHAROFFSET              8
#define TRANSMITTED             5

/* Syscall mnemonic values */
#define SEND                    1
#define RECV                    2

/* service request values */
#define CREATEBROTHER   1
#define CREATESON       2
#define TERMINATE       3
#define SPECPRGMGR      4
#define SPECTLBMGR      5
#define SPECSYSMGR      6
#define GETCPUTIME      7
#define WAITFORCLOCK    8
#define WAITFORIO       9

/* specific message values */
#define ANYMESSAGE		0
#define TRAPCONTINUE		1
#define TRAPTERMINATE		0
 
#define CREATENOGOOD            -1      /* For thread creation failure */
#define MSGNOGOOD		-1     	/* For message send failure */



/* operations */
#define	MIN(A,B)	((A) < (B) ? A : B)
#define MAX(A,B)	((A) < (B) ? B : A)
#define	ALIGNED(A)	(((unsigned)A & 0x3) == 0)

/* Useful operations */
#define STCK(T) ((T) = ((* ((cpu_t *) TODLOADDR)) / (* ((cpu_t *) TIMESCALEADDR))))

#define MsgSend(code,dest,payload) (SYSCALL((unsigned int) (code),(unsigned int) (dest),(unsigned int) (payload),0))

#define MsgRecv(code,source,reply) (((tcb_t *) SYSCALL((unsigned int) (code),(unsigned int) (source),(unsigned int) (reply),0)))

#endif
